@@ -59,7 +59,6 @@ def _plot_bucket_thumbnails(bucket_stats: List[Dict[str, Any]], viz_data: Dict)
         import matplotlib.pyplot as plt
         import numpy as np
         from PIL import Image
-        import random
     except ImportError:
         print("matplotlib not available - skipping thumbnail grids")
         return
@@ -74,17 +73,21 @@ def _plot_bucket_thumbnails(bucket_stats: List[Dict[str, Any]], viz_data: Dict)
     bucket_assignments = viz_data['bucket_assignments']
     all_paths = viz_data['all_paths']
     
-    # Calculate grid layout for subplots
+    # Calculate square grid layout for subplots
     n_buckets = len(sorted_buckets)
-    cols = min(4, n_buckets)
+    cols = int(np.ceil(np.sqrt(n_buckets)))
     rows = int(np.ceil(n_buckets / cols))
     
     # Create figure
     fig, axes = plt.subplots(rows, cols, figsize=(cols * 5, rows * 5))
-    if n_buckets == 1:
-        axes = [axes]
+    
+    # Ensure axes is always a 2D array for consistent access
+    if rows == 1 and cols == 1:
+        axes = np.array([[axes]])
     elif rows == 1:
         axes = axes.reshape(1, -1)
+    elif cols == 1:
+        axes = axes.reshape(-1, 1)
     
     # Process each bucket
     for bucket_idx, bucket_data in enumerate(sorted_buckets):
@@ -101,8 +104,8 @@ def _plot_bucket_thumbnails(bucket_stats: List[Dict[str, Any]], viz_data: Dict)
             if not images:
                 return np.ones((300, 300, 3), dtype=np.uint8) * 220  # Gray placeholder
             
-            # Randomly sample up to 100 images
-            sample_images = random.sample(images, min(len(images), max_images))
+            # Take first 100 images (not random) for consistent display
+            sample_images = images[:max_images]
             
             # Create 10x10 grid (300x300 pixels, each thumbnail 30x30)
             grid_img = np.ones((300, 300, 3), dtype=np.uint8) * 255  # White background
@@ -140,29 +143,18 @@ def _plot_bucket_thumbnails(bucket_stats: List[Dict[str, Any]], viz_data: Dict)
         # Calculate subplot position
         row = bucket_idx // cols
         col = bucket_idx % cols
-        
-        if rows > 1:
-            ax = axes[row, col]
-        else:
-            ax = axes[col]
+        ax = axes[row, col]
         
         # Create and display thumbnail grid
         grid_img = create_bucket_grid(bucket_images)
         ax.imshow(grid_img)
-        
-        # Set title with just bucket number
-        ax.set_title(f"Bucket {bucket_idx + 1}", fontsize=14, pad=10)
         ax.axis('off')
     
     # Hide unused subplots
     for i in range(n_buckets, rows * cols):
         row = i // cols
         col = i % cols
-        if rows > 1:
-            axes[row, col].axis('off')
-        else:
-            if i < len(axes):
-                axes[i].axis('off')
+        axes[row, col].axis('off')
     
     plt.suptitle('Bucket Thumbnails: Visual Similarity Groups (10x10 grid, max 100 images per bucket)', 
                  fontsize=14, y=0.95)
@@ -240,8 +232,8 @@ def sample_diverse(
     Preserves maximum diversity by ensuring representation from all visual groups.
     
     Strategy:
-    1. Compute quick hashes for all images
-    2. Group similar images into buckets (~16 visual groups)
+    1. Compute multi-dimensional visual features (structure, color, brightness)
+    2. Group similar images into buckets based on visual similarity
     3. Diversity-first: Sample from every bucket to preserve visual variety, then fill largest buckets
     
     Args:
@@ -291,9 +283,34 @@ def sample_diverse(
             with Image.open(path) as img:
                 if img.mode != 'RGB':
                     img = img.convert('RGB')
-                # Use dhash for speed (faster than phash)
-                hash_val = imagehash.dhash(img, hash_size=hash_size)
-                return str(path), hash_val
+                
+                # Compute multiple features for better visual similarity
+                # 1. DHash for structural/edge information
+                dhash = imagehash.dhash(img, hash_size=hash_size)
+                
+                # 2. Average hash for overall brightness/contrast
+                ahash = imagehash.average_hash(img, hash_size=hash_size//2)
+                
+                # 3. Color histogram features (simplified)
+                img_small = img.resize((32, 32), Image.Resampling.LANCZOS)
+                img_array = np.array(img_small)
+                
+                # Color variance (low = grayscale/monochrome, high = colorful)
+                r, g, b = img_array[:,:,0], img_array[:,:,1], img_array[:,:,2]
+                color_variance = np.var([np.mean(r), np.mean(g), np.mean(b)])
+                
+                # Brightness level
+                brightness = np.mean(img_array)
+                
+                # Combine all features
+                combined_features = {
+                    'dhash': dhash,
+                    'ahash': ahash,
+                    'color_variance': color_variance,
+                    'brightness': brightness
+                }
+                
+                return str(path), combined_features
         except:
             return str(path), None
     
@@ -309,31 +326,54 @@ def sample_diverse(
             futures_iter = futures
         
         for future in futures_iter:
-            path, hash_val = future.result()
-            if hash_val is not None:
+            path, features = future.result()
+            if features is not None:
                 valid_paths.append(path)
-                hashes.append(hash_val)
+                hashes.append(features)
     
     n_valid = len(valid_paths)
     
     if target_count >= n_valid:
         return valid_paths
     
-    # Step 2: Create diversity grid using hash bits
+    # Step 2: Create diversity grid using multi-dimensional features
     if show_progress:
-        print("Creating diversity grid...")
+        print("Creating diversity grid with enhanced visual features...")
     
-    # Convert hashes to bit arrays for bucketing
-    hash_arrays = np.array([np.array(h.hash).flatten() for h in hashes])
-    
-    # Use first few bits to create buckets (trade-off between diversity and speed)
-    n_bucket_bits = min(4, hash_size)  # Use 4 bits = 16 buckets max
     bucket_keys = []
     
-    for h in hash_arrays:
-        # Create bucket key from first few bits
-        key = tuple(h.flatten()[:n_bucket_bits] > 0)
-        bucket_keys.append(key)
+    for features in hashes:
+        dhash = features['dhash']
+        ahash = features['ahash']
+        color_var = features['color_variance']
+        brightness = features['brightness']
+        
+        # Convert dhash to bits (structural features)
+        dhash_bits = np.array(dhash.hash).flatten()
+        ahash_bits = np.array(ahash.hash).flatten()
+        
+        # Use fewer bits from dhash for structure (reduce from 4 to 3)
+        total_dhash_bits = len(dhash_bits)
+        if total_dhash_bits >= 3:
+            dhash_indices = [0, total_dhash_bits//3, 2*total_dhash_bits//3]
+            structure_bits = tuple(dhash_bits[i] > 0 for i in dhash_indices)
+        else:
+            structure_bits = tuple(dhash_bits > 0)
+        
+        # Use only 1 bit from ahash for basic brightness pattern
+        brightness_bit = (ahash_bits[0] > 0,) if len(ahash_bits) >= 1 else ()
+        
+        # Simplified color variance (0=grayscale, 1=colored)
+        color_bucket = 0 if color_var < 100 else 1  # Binary: grayscale vs color
+        
+        # Simplified brightness (0=dark, 1=bright)
+        bright_bucket = 0 if brightness < 128 else 1  # Binary: dark vs bright
+        
+        # Combine all features into bucket key
+        # 3 structure + 1 brightness + 1 color + 1 overall_brightness = 6 dimensions
+        # Max buckets: 2^3 × 2^1 × 2 × 2 = 8 × 2 × 2 × 2 = 64 buckets
+        bucket_key = structure_bits + brightness_bit + (color_bucket, bright_bucket)
+        bucket_keys.append(bucket_key)
     
     # Group images by bucket
     buckets = {}
@@ -351,7 +391,12 @@ def sample_diverse(
         buckets[key] = [pair[1] for pair in sorted_pairs]
     
     if show_progress:
-        print(f"Grouped into {len(buckets)} diversity buckets with natural sorting")
+        print(f"Grouped into {len(buckets)} diversity buckets with enhanced visual features")
+        print(f"Features: structure (3 bits) + brightness pattern (1 bit) + color type (2 levels) + brightness level (2 levels)")
+        
+        # Show bucket size distribution
+        bucket_sizes = [len(indices) for indices in buckets.values()]
+        print(f"Bucket sizes: min={min(bucket_sizes)}, max={max(bucket_sizes)}, avg={sum(bucket_sizes)/len(bucket_sizes):.1f}")
     
     # Step 3: Diversity-preserving selection
     if show_progress:
